/**
 * $Id: Bag.java,v 1.4 2006/07/07 11:38:53 uwe Exp $
 */
package de.fhw.data.bag;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.list.List;
import de.fhw.data.util.Fingerprint;
import de.fhw.data.util.Invariant;
import de.fhw.data.util.Profile;

/**
 * The interface for all persistent Bag implementations
 * 
 */
public interface Bag<E>
extends Invariant, Profile, Fingerprint{
	
	/**
	 * test on empty Bag
	 */
	boolean isEmpty();
	
	/**
	 * 
	 * @return number of elements in the Bag
	 */
	int card();
	
	/**
	 * # of occurences of an element in a bag
	 */
	int occurences(E e);

	/**
	 * search for an element
	 */
	boolean contains(E e);

	/**
	 * test whether this bag contains a bag b2
	 * <p>
	 * <code>b1.contains(b2)</code> is equivalent to
	 * <code>b2.isSubbagOf(b1)</code>
	 */
	boolean contains(Bag<E> b2);

	/**
	 * test whether this bag is a subbag or equal to bag b2
	 */
	boolean isSubbagOf(Bag<E> b2);
	
	/**
	 * test whether this bag is equal to another bag
	 */
	boolean isEqualTo(Bag<E> b2);
	
	/**
	 * build a new empty bag
	 */
	Bag <E> empty();

	/**
	 * build a single element bag
	 */
	Bag <E> single(E e);

	/**
	 * build a single bag with a single element occuring multiple times
	 */
	Bag <E> single(E e, int occ);

	/**
	 * insert an element
	 */
	Bag <E> insert(E e);

	/**
	 * insert an element multiple times
	 */
	Bag <E> insert(E e, int occ);
	
	/**
	 * remove an element
	 */
	Bag <E> remove(E e);

	/**
	 * remove multiple occurences of an element
	 */
	Bag <E> remove(E e, int occ);
	
	/**
	 * choose an arbitray element from a none empty bag,
	 * if the bag is empty, an IllegalArgumentException is thrown
	 */
	E choose();

	/**
	 * Only supported for ordered bags.
	 * If the bag is empty, an IllegalArgumentException is thrown.
	 * If the bag isn't ordered, an UnsupportedOperationException is thrown
	 * 
	 * @return the minimum element in a none empty bag.
	 */
	E findMin();
	
	/**
	 * Only supported for ordered bags
	 * If the bag is empty, an IllegalArgumentException is thrown.
	 * If the bag isn't ordered, an UnsupportedOperationException is thrown
	 * 
	 * @return the maximum element in a none empty bag
	 */
	E findMax();

	/**
	 * remove the minimum element,
	 * if the bag is empty the operation returns the empty bag
	 * 
	 * Only supported for ordered bags
	 */
	Bag <E> removeMin();
	
	/**
	 * remove the maximum element,
	 * if the bag is empty the operation returns the empty bag
	 * 
	 * Only supported for ordered bags
	 */
	Bag <E> removeMax();

	/**
	 * union of 2 bags
	 */	
	Bag <E> union(final Bag <E> b2);
	
	/**
	 * difference of 2 bags
	 */
	Bag <E> difference(final Bag <E> b2);
	
	/**
	 * intersection of 2 bags
	 */
	Bag <E> intersection(final Bag <E> b2);
	
	/**
	 * simple version of map: maps all elements,
	 * but works with a single bag type
	 */
	Bag <E> map(final Fct<E,E> mf);
	
	/**
	 * map all elements of a bag and collect the results in a new bag
	 */
	<E2>
	Bag <E2> map(final Fct<E, E2> mf, Bag<E2> proto);
	
	/**
	 * filter the elements of a bag
	 * and remove all elements, for which the predicate doesn not hold
	 * @param filter the predicate
	 * @return the new possibly smaller bag
	 */
	Bag <E> filter(final Fct<E,Boolean> filter);
	
	/**
	 * fold all elements of a bag into a new result
	 */
	<R>
	R fold(final Fct2<E,R,R> f, R init);
	
	/**
	 * map all elements of a bag into a result type
	 * and combine these results, if the bag is empty,
	 * the 3. parameter is taken as the result
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param init the initial value for building the result,
	 *  it's theresult for the empty bag
	 * @return the combined result
	 */
	<R>
	R foldMap(final Fct2<R,R,R> combine, final Fct<E,R> map, R init);
	
	/**
	 * map all elements of a bag into a result type
	 * and combine these results, emptyRes is the result if the bag is empty,
	 * otherwise this value is not used, especially for
	 * one element bags the result is the mapping of the single element, not the
	 * combination of the element and emptyRes.
	 * 
	 * this allows building of lists like "e1,e2,...,en" and "e1"
	 * where foldMap would build "e1,e2,...,en," and "e1,"
	 * 
	 * example application is the general toString method in de.fhw.data.bag.BagImpl
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param emptyRes the result for the empty bag
	 * @return the combined result
	 */
	<R>
	R fold1Map(final Fct2<R,R,R> combine, final Fct<E,R> map, R emptyRes);
	
	/**
	 * test, whether for all elements in a bag a predicate p holds,
	 * generalisation of logical and
	 */
	Boolean all(Fct<E,Boolean> p);
	
	/**
	 * test, whether for some elements in a bag a predicate p holds,
	 * generalisation of logical or
	 */
	Boolean any(Fct<E,Boolean> p);

	/**
	 * conversion into a list
	 */
	List<E> toList(List<E> proto);
}
