/**
 * $Id: Predef.java,v 1.2 2006/06/14 05:55:38 uwe Exp $
 */
package de.fhw.data.function;

import de.fhw.data.map.Map;
import de.fhw.data.set.Set;

/**
 * <p>
 * collection of predefined function objects
 * and function generators.
 * </p><p>
 * These classes contain only static fields and
 * static methods
 * </p><p>
 * The generic function generators normally must be used
 * with explicit variable declarations, otherwise, e.g. in fold and map operations,
 * the compiler issues type check errors.
 * </p><p>
 * Example:
 * </p><p> <code>Fct&lt;Integer,Integer&gt; id = Prelude.id(); ... s.map(id) ...</code>
 * </p><p>
 * instead of
 * </p><p> <code>... s.map(Prelude.id()) ...</code> 
 * </p>
 * 
 * @see Set#map
 * @see Set#fold
 * @see Set#foldMap
 * @see Set#filter
 * @see Map#map
 * @see Map#fold
 * @see Map#foldMap
 */

public class Predef {
	
	/**
	 * predefined function objects
	 * for boolean operations
	 *
	 */
	public static class Bool {
		/**
		 * logical and
		 */
		public static final Fct2<Boolean,Boolean,Boolean> and =
			new Fct2<Boolean,Boolean,Boolean>() {
			public Boolean apply(Boolean x1, Boolean x2) {
				return x1 && x2;
			}
		};
		/**
		 * logical or
		 */
		public static final Fct2<Boolean,Boolean,Boolean> or =
			new Fct2<Boolean,Boolean,Boolean>() {
			public Boolean apply(Boolean x1, Boolean x2) {
				return x1 || x2;
			}
		};
		/**
		 * implication
		 */
		public static final Fct2<Boolean,Boolean,Boolean> implies =
			new Fct2<Boolean,Boolean,Boolean>() {
			public Boolean apply(Boolean x1, Boolean x2) {
				return ! x1 || x2;
			}
		};
		/**
		 * logical not
		 */
		public static final Fct<Boolean,Boolean> not =
			new Fct<Boolean,Boolean>() {
			public Boolean apply(Boolean x1) {
				return ! x1;
			}
		};
	}
	
	/**
	 * predefined function objects
	 * for Integer operations
	 *
	 */
	public static class Int {
		/**
		 * integer addition
		 */
		public static final
		Fct2<Integer,Integer,Integer> plus =
			new Fct2<Integer,Integer,Integer>() {
			public Integer apply(Integer x1, Integer x2) {
				return x1 + x2;
			}
		};
		
		/**
		 * integer subtraction
		 */
		public static final
		Fct2<Integer,Integer,Integer> minus =
			new Fct2<Integer,Integer,Integer>() {
			public Integer apply(Integer x1, Integer x2) {
				return x1 - x2;
			}
		};
		
		/**
		 * integer multiplication
		 */
		public static final
		Fct2<Integer,Integer,Integer> mult	 =
			new Fct2<Integer,Integer,Integer>() {
			public Integer apply(Integer x1, Integer x2) {
				return x1 * x2;
			}
		};
		
		/**
		 * integer division
		 */
		public static final
		Fct2<Integer,Integer,Integer> div =
			new Fct2<Integer,Integer,Integer>() {
			public Integer apply(Integer x1, Integer x2) {
				return x1 / x2;
			}
		};
		/**
		 * integer negation
		 */
		public static final
		Fct<Integer,Integer> negate =
			new Fct<Integer,Integer>() {
			public Integer apply(Integer x1) {
				return 0 - x1;
			}
		};
		
		/**
		 * integer comparison
		 */
		public static final
		Fct2<Integer,Integer,Integer> cmp =
			new Fct2<Integer,Integer,Integer>() {
			public
			Integer apply(Integer x1, Integer x2) {
				if (x1 > x2)
					return 1;
				if (x1 < x2)
					return -1;
				return 0;
			}
		};
	}
	/**
	 * predefined function objects
	 * for Set operations
	 *
	 */
	public static class Sets {
		/**
		 * set union
		 */
		public static
		<E>
		Fct2<Set<E>,Set<E>,Set<E>> union() {
			return
			new Fct2<Set<E>,Set<E>,Set<E>>() {
				public
				Set<E> apply(Set<E> x1,Set<E> x2) {
					return x1.union(x2);
				}
			};
		}

		/**
		 * set intersection
		 */
		public static
		<E>
		Fct2<Set<E>,Set<E>,Set<E>> intersection() {
			return
			new Fct2<Set<E>,Set<E>,Set<E>>() {
				public
				Set<E> apply(Set<E> x1,Set<E> x2) {
					return x1.intersection(x2);
				}
			};
		}
		
		/**
		 * set difference
		 */
		public static
		<E>
		Fct2<Set<E>,Set<E>,Set<E>> difference() {
			return
			new Fct2<Set<E>,Set<E>,Set<E>>() {
				public
				Set<E> apply(Set<E> x1,Set<E> x2) {
					return x1.difference(x2);
				}
			};
		}
		
		/**
		 * function for building single element sets
		 * @param <E> the element type
		 * @param proto the prototype for the set to be build
		 * @return the function object
		 */
		public static
		<E>
		Fct<E,Set<E>> singleSet(final Set<E> proto) {
			return
			new Fct<E,Set<E>>() {
				public
				Set<E> apply(E e) {
					return proto.single(e);
				}
			};
		}
	}
	
	/**
	 * predefined function objects
	 * for String operations
	 */
	public static class Str {
		/**
		 * string concatenation
		 */
		public static final Fct2<String,String,String> concat =
			new Fct2<String,String,String>() {
			public String apply(String x1, String x2) {
				return x1 + x2;
			}
		};
		
		/**
		 * string concatenation with a delimiter
		 * 
		 * @param del the delimiter string
		 * @return the function object for joining two strings
		 */
		public static Fct2<String,String,String> join(final String del) {
			return
			new Fct2<String,String,String>() {
				public String apply(String x1, String x2) {
					return x1 + del + x2;
				}
			};
		}
		
		/**
		 * String comparison
		 */
		public static final
		Fct2<String,String,Integer> cmp =
			new Fct2<String,String,Integer>() {
			public
			Integer apply(String x1, String x2) {
				int r = x1.compareTo(x2);
				if (r == 0) return 0;
				if (r > 0) return 1;
				return -1;
			}
		};

	}
	/**
	 * static method for creation of
	 * conversion functions into strings
	 * 
	 *  implemented with toString
	 *  
	 * @param <X> the element type
	 */
	public static
	<X>
	Fct<X,String> show() {
		return
		new Fct<X,String>() {
			public
			String apply(X x1) {
				return
				x1.toString();
			}
		};
	}

	/**
	 * identity construction function
	 * 
	 * @return the identity for a given argument type
	 */
	public static
	<X>
	Fct<X,X> id() {
		return
		new Fct<X,X>() {
			public
			X apply(X x1) {
				return x1;
			}
		};
	}
	/**
	 * const function generator
	 * 
	 * @param <X> the argument type of the argument to be ignored 
	 * @param <R> the result type
	 * @param res the result of the apply function
	 * @return the const function object
	 */
	public static
	<X,R>
	Fct<X,R> constant(final R res) {
		return
		new Fct<X,R>() {
			public
			R apply(X x1) {
				return res;
			}
		};
	}
	
	/**
	 * function composition
	 * 
	 * @param <X> the type of the argument of the 1. function
	 * @param <X2> the type of the result of the 1. function and the arugment of the 2.
	 * @param <R> the result of the 2. function
	 * @param f1 the 1. function object
	 * @param f2 the 2. function object
	 * @return a function object representing the function composition
	 */
	public static
	<X,X2,R>
	Fct<X,R> compose(final Fct<X,X2>f1, final Fct<X2,R> f2) {
		return
		new Fct<X,R>() {
			public
			R apply(X x1) {
				return
				f2.apply(f1.apply(x1));
			}
		};
	}

	/**
	 * function composition
	 * 
	 * @param <X1> the type of the 1. argument of the 1. function
	 * @param <X2> the type of the 2. argument of the 1. function
	 * @param <Y> the type of the result of the 1. function and the arugment of the 2.
	 * @param <R> the result of the 2. function
	 * @param f1 the 1. function object
	 * @param f2 the 2. function object
	 * @return a function object representing the function composition
	 */
	public static
	<X1,X2,Y,R>
	Fct2<X1,X2,R> compose(final Fct2<X1,X2,Y>f1, final Fct<Y,R> f2) {
		return
		new Fct2<X1,X2,R>() {
			public
			R apply(X1 x1, X2 x2) {
				return
				f2.apply(f1.apply(x1,x2));
			}
		};
	}
	
	/**
	 * generic comparator creation
	 *  
	 * @param <X> the type of the values to be compared,
	 *  this type must implement the java.lang.Comparable interface
	 *  
	 * @return the comparator object
	 */
	public static
	<X extends Comparable<? super X>>
	Fct2<X,X,Integer> cmp() {
		return
		new Fct2<X,X,Integer>() {
			public
			Integer apply(X x1, X x2) {
				int res = x1.compareTo(x2);
				if (res == 0) return 0;
				if (res > 0) return 1;
				return -1;
			}
		};
	}

	/**
	 * generic comparator creation for reverse ordering
	 *  
	 * @param <X> the type of the values to be compared,
	 *  this type must implement the java.lang.Comparable interface
	 *  
	 * @return the comparator object for the reverse ordering
	 */

	public static
	<X extends Comparable<? super X>>
	Fct2<X,X,Integer> cmpReverse() {
		/*
		 * this version does not compile with javac
		 * 	
		Fct2<X,X,Integer> c = cmp();
		return
		compose(c,Int.negate);
		*/
		
		return
		new Fct2<X,X,Integer>() {
			public
			Integer apply(X x1, X x2) {
				int res = x1.compareTo(x2);
				if (res == 0) return 0;
				if (res > 0) return 1;
				return -1;
			}
		};
	}


	/*
	 * an old version of cmpReverse
	 * 

	public static
	<X extends Comparable<? super X>>
	Fct2<X,X,Integer> cmpReverse() {
		return
		new Fct2<X,X,Integer>() {
			public
			Integer apply(X x1, X x2) {
				Fct2<X,X,Integer> c = cmp();
				return
				c.apply(x2,x1);
			}
		};
	}
	*/
}
