/**
 * $Id: List.java,v 1.3 2006/07/07 11:38:53 uwe Exp $
 */
package de.fhw.data.list;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.tuple.Pair;
import de.fhw.data.util.Fingerprint;
import de.fhw.data.util.Invariant;
import de.fhw.data.util.Profile;

/**
 * The interface for all persistent list implementations
 * 
 */
public interface List<E>
extends Invariant, Profile, Fingerprint {
	
	/**
	 * test on empty list
	 */
	boolean isEmpty();
	
	/**
	 * 
	 * @return the length of the list
	 */
	int length();
		
	/**
	 * test whether this list is equal to another list
	 */
	boolean isEqualTo(List<E> l2);

	/**
	 * build a new empty list
	 */
	List<E> empty();

	/**
	 * build a single element list
	 */
	List<E> single(E e);

	/**
	 * add an element at the top of the list
	 */
	List<E> cons(E e);
	
	/**
	 * add an element at the end of the list
	 */
	List<E> append(E e);
	
	/**
	 * access the head of a none empty list,
	 * if the list is empty, an IndexOutOfBoundsException is thrown
	 */
	E head();

	/**
	 * access the last element of a none empty list,
	 * if the list is empty, an IndexOutOfBoundsException is thrown
	 */
	E last();

	/**
	 * access the i. element of a list
	 * IndexOutOfBoundsException is raised for illegal index values
	 */
	E at(int i);
	
	/**
	 * drop head of a none empty list,
	 * if the list is empty, an IllegalArgumentException is thrown
	 */
	List<E> tail();

	/**
	 * drop the last element of a none empty list,
	 * if the list is empty, an IllegalArgumentException is thrown
	 */
	List<E> init();

	/**
	 * take the first i elements of a list
	 * if the list contains less elements, the whole list
	 * is the result. if i is negative, the result is the empty list.
	 */
	List<E> take(int i);

	/**
	 * drop the first i elements of a list.
	 * if the list contains less elements, the result is the empty list,
	 * if i is negative the result is the whole list
	 */
	List<E> drop(int i);

	/**
	 * split a list at a position i and return a pair of results
	 * <p>
	 * splitAt is a combination of take and drop
	 */ 
	Pair<List<E>,List<E>> splitAt(int i);
	
	/**
	 * insert an element in front of index i
	 * <p>
	 * the list is split at position i, the result is build
	 * by concatenating the take part with the drop part consed with the element v
	 * <p>
	 * Haskell Spec: <code>insertAt i v l = take i l ++ [v] ++ drop i l</code>
	 */
	List<E> insertAt(int i, E v);

	/**
	 * remove the element at position i
	 * <p>
	 * the list is split at position i, the result is build
	 * by concatenating the take part with the tail of the drop part.
	 * An IndexOutOfBoundsException can be raised for illegal indices
	 * <p>
	 * Haskell Spec: <code>removeAt i l = take i l ++ drop (i+1) l</code>
	 */
	List<E> removeAt(int i);
	
	/**
	 * change an element at an index i
	 * <p>
	 * the list is split at position i, the result is build
	 * by concatenating the take part with the tail of the drop part consed with the element v
	 * <p>
	 * Haskell Spec: <code>updateAt i v l = take i l ++ [v] ++ drop (i+1) l</code>
	 * <p>
	 * <code>l.updateAt(i,v)</code> is equivalent to
	 * <code>l.removeAt(i).insertAt(i,v)</code>
	 */
	List<E> updateAt(int i, E v);
		
	/**
	 * concatenation of 2 lists
	 */	
	List<E> concat(final List<E> l2);
	
	/**
	 * reverse a list
	 */
	List<E> reverse();
		
	/**
	 * simple version of map: maps all elements,
	 * but works with a single list type
	 */
	List<E> map(final Fct<E,E> mf);
	
	/**
	 * map all elements of a list and collect the results in a list of another type
	 */
	<E2>
	List<E2> map(final Fct<E, E2> mf, List<E2> proto);
	
	/**
	 * filter the elements of a list
	 * and remove all elements, for which the predicate doesn't hold
	 * 
	 * @param filter the predicate
	 * @return the new possibly smaller list
	 */
	List<E> filter(final Fct<E,Boolean> filter);

	/**
	 * fold all elements of a list into a new result,
	 * folding is done from the left
	 */
	<R>
	R foldl(final Fct2<R,E,R> f, R init);
	
	/**
	 * fold all elements of a list into a new result,
	 * folding is done from the right
	 */
	<R>
	R foldr(final Fct2<E,R,R> f, R init);
	
	/**
	 * map all elements of a list into a result type
	 * and combine these results, if the list is empty,
	 * the 3. parameter is taken as the result
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param init the initial value for building the result,
	 *  it's theresult for the empty list
	 * @return the combined result
	 */
	<R>
	R foldrMap(final Fct2<R,R,R> combine, final Fct<E,R> map, R init);
	
	/**
	 * map all elements of a list into a result type
	 * and combine these results, emptyRes is the result if the list is empty,
	 * otherwise this value is not used, especially for
	 * one element sets the result is the mapping of the single element, not the
	 * combination of the element and emptyRes.
	 * 
	 * this allows building of lists like "e1,e2,...,en" and "e1"
	 * where foldMap would build "e1,e2,...,en," and "e1,"
	 * 
	 * example application is the general toString method in de.fhw.data.list.SetImpl
	 * 
	 * @param <R> the result type 
	 * @param combine the combining function
	 * @param map the element mapping
	 * @param emptyRes the result for the empty list
	 * @return the combined result
	 */
	<R>
	R foldr1Map(final Fct2<R,R,R> combine, final Fct<E,R> map, R emptyRes);

	/**
	 * combine two lists elementwise
	 * @param zip the function for computing the result element
	 * @param l2 the 2. list
	 * @return a list with the length of the shorter list, where
	 * all elements are computed with the zip function
	 */
	List<E> zipWith(Fct2<E,E,E> zip, List<E> l2);

	/**
	 * general form of zipWith with different types of lists
	 * 
	 * @see #zipWith(Fct2, List)
	 * @see #map(Fct, List)
	 */
	<E2,R>
	List<R> zipWith(final Fct2<E,E2,R> zip, List<E2> l2, List<R> proto);
	
	/**
	 * test, whether for all elements in a list a predicate p holds,
	 * generalisation of logical and
	 */
	Boolean all(Fct<E,Boolean> p);
	
	/**
	 * test, whether for some elements in a list a predicate p holds,
	 * generalisation of logical or
	 */
	Boolean any(Fct<E,Boolean> p);
}
