/**
 * $Id: ListImpl.java,v 1.4 2006/07/07 11:38:54 uwe Exp $
 */
package de.fhw.data.list;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Predef;
import de.fhw.data.tuple.Pair;


/**
 * The interface for all persistent list implementations
 * 
 */
public abstract
class ListImpl<E>
implements List<E> {
	
	/**
	 * test on empty list implemented with length
	 */
	public
	boolean isEmpty() {
		return
		length() == 0;
	}
	
	/**
	 * 
	 * length of the list implemented with fold
	 */
	public
	int length() {
		return
		foldr(new Fct2<E,Integer,Integer>() {
			public
			Integer apply(E e, Integer r) {
				return
				r + 1;
			}
		}
		, 0
		);
	}
		
	/**
	 * test whether this list is equal to another list
	 */
	public
	boolean isEqualTo(List<E> l2) {
		for (int i=0; i<length(); ++i) {
			if (!at(i).equals(l2.at(i)))
				return false;
		}
		return true;
	}

	/**
	 * build a single element list with empty and cons
	 */
	public
	List <E> single(E e) {
		return empty().cons(e);
	}

	/**
	 * append implemeted with concat and single
	 */
	public
	List <E> append(E e) {
		return concat(single(e));
	}
	
	/**
	 * head implemeted with at
	 */
	public
	E head() {
		return at(0);
	}

	/**
	 * last implemented with at and length
	 */
	public
	E last() {
		return at(length() - 1);
	}

	/**
	 * at implemented with head and tail,
	 * <p>
	 * at least one method of head and at must be reimplemented
	 * else there is a cyclic definition
	 */
	public
	E at(int i) {
		if (i == 0)
			return head();
		
		return tail().at(i-1);
	}
	
	/**
	 * tail implemented with removeAt
	 */
	public
	List <E> tail() {
		return removeAt(0);
	}

	/**
	 * init implemented with removeAt and length
	 */
	public
	List <E> init() {
		return removeAt(length() - 1);
	}

	/**
	 * take implemented with splitAt
	 */
	public
	List<E> take(int i) {
		return splitAt(i).fst;
	}

	/**
	 * drop implemented with splitAt
	 */
	public
	List<E> drop(int i) {
		return splitAt(i).snd;
	}

	/**
	 * splitAt implemented with take and drop
	 */
	public
	Pair<List<E>,List<E>> splitAt(int i) {
		List<E> l1 = take(i);
		List<E> l2 = drop(i);
		return new Pair<List<E>,List<E>>(l1,l2);
	}
	
	/**
	 * insertAt implemented with splitAt, cons and concat
	 */
	public
	List<E> insertAt(int i, E v) {
		Pair<List<E>,List<E>> p = splitAt(i);
		return p.fst.concat(p.snd.cons(v));
	}

	/**
	 * remove implemented with splitAt, tail and concat
	 */
	public
	List<E> removeAt(int i) {
		Pair<List<E>,List<E>> p = splitAt(i);
		return p.fst.concat(p.snd.tail());
	}
	
	/**
	 * update implemented with splitAt, tail, cons and concat
	 */
	public
	List<E> updateAt(int i, E v){
		Pair<List<E>,List<E>> p = splitAt(i);
		return p.fst.concat(p.snd.tail().cons(v));
	}
	
	/**
	 * concat implemented with fold and append
	 * <p>
	 * cyclic default implementations: one of append or concat
	 * must be implemented
	 */
	public
	List <E> concat(final List <E> l2) {
		Fct2<E,List<E>,List<E>> app =
			new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E v, List<E> r) {
				return r.append(v);
			}
		};
		
		return l2.foldr(app,this);
	}

	/**
	 * reverse implemented with fold and cons
	 */
	public
	List<E> reverse() {
		Fct2<E,List<E>,List<E>> cons =
			new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E v, List<E> r) {
				return r.concat(r.single(v));
			}
		};
		return foldr(cons,empty());
	}
		
	/**
	 * map implemented with foldr and cons,
	 */
	public
	List <E> map(final Fct<E,E> mf) {
		
		Fct2<E,List<E>,List<E>> f
		= new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E e, List<E> res) {
				return
				res.cons(mf.apply(e));
			}
		};
		return
		foldr(f,empty());
	}
	
	/**
	 * map implemeted with foldr and cons,
	 */
	public
	<E2>
	List <E2> map(final Fct<E, E2> mf, List<E2> proto) {

		Fct2<E,List<E2>,List<E2>> f
		= new Fct2<E,List<E2>,List<E2>>() {
			public
			List<E2> apply(E e, List<E2> res) {
				return
				res.cons(mf.apply(e));
			}
		};
		return
		foldr(f,proto.empty());
		
	}
	
	/**
	 * filter implemeted with fold
	 */
	public
	List <E> filter(final Fct<E,Boolean> filter) {
		Fct2<E,List<E>,List<E>> f
		= new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E e, List<E> res) {
				return
				filter.apply(e)
				? res.append(e) : res;
			}
		};
		return
		foldr(f,empty());
		
	}

	/**
	 * zipWith implemented with length and at,
	 * efficient only if at is efficient
	 */
	public
	List<E> zipWith(Fct2<E,E,E> zip, List<E> l2) {
		int i = Math.min(length(),l2.length());
		List<E> res = empty();
		while (i != 0) {
			--i;
			res = res.cons(zip.apply(at(i),l2.at(i)));
		}
		return res;
	}

	/**
	 * general zipWith implemented with length and at
	 * efficient only if at is efficient
	 */
	public
	<E2,R>
	List<R> zipWith(final Fct2<E,E2,R> zip, List<E2> l2, List<R> proto) {
		int i = Math.min(length(),l2.length());
		List<R> res = proto.empty();
		while (i != 0) {
			--i;
			res = res.cons(zip.apply(at(i),l2.at(i)));
		}
		return res;
	}
	
	
	/**
	 * all quantor implemented with fold
	 */
	public
	Boolean all(Fct<E,Boolean> p) {
		return
		foldrMap(Predef.Bool.and,p,true);
	}
	
	/**
	 * exists quantor implemented with fold
	 */
	public
	Boolean any(Fct<E,Boolean> p) {
		return
		foldrMap(Predef.Bool.or,p,false);
	}

	/**
	 * default invariant: true
	 */
	public
	boolean inv() {
		return true;
	}

	@Override
	public
	boolean equals(Object o2) {
		if (o2 instanceof List)
			return this.isEqualTo((List<E>)o2);
		return false;
	}

	/**
	 * toString is implemented with isEmpty and fold1Map
	 */
	@Override
	public
	String toString() {
		if (isEmpty())
			return "[]";
		return
		"["
		+ foldr1Map(Predef.Str.join(", "), Predef.<E>show(), "")
		+ "]";
	}
}
