/**
 * $Id: Map.java,v 1.6 2006/02/24 21:21:01 uwe Exp $
 */
package de.fhw.data.map;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Fct3;
import de.fhw.data.set.Set;
import de.fhw.data.tuple.Pair;
import de.fhw.data.util.Fingerprint;
import de.fhw.data.util.Invariant;
import de.fhw.data.util.Profile;
import de.fhw.data.list.*;
/**
 * common map interface
 *
 * @param <K> the key type
 * @param <V> the attribute value type
 */
public interface Map<K, V>
extends Invariant, Profile, Fingerprint {
	
	/**
	 * test on empty map
	 */
	boolean isEmpty();
	
	/**
	 * number of key value pairs in a map
	 */
	int card();
	
	/**
	 * test whether a key occurs in a map
	 */
	boolean contains(K k);
	
	/**
	 * test whether this map is equal to another map
	 */
	boolean isEqualTo(Map<K,V> s2);
	
	/**
	 * access the value associated with a key.
	 * if the key is not found, the null reference is returned
	 */
	V getValue(K k);

	/**
	 * access the value associated with a key,
	 * save variant of getValue:
	 * If the key is not found, an IndexOutOfBoundsException is thrown
	 */
	V at(K k);
	
	/**
	 * generate a new empty map of the same type as the object
	 */	
	Map <K, V> empty();
	
	/**
	 * insert a key value pair,
	 * if the key already exists, insert does an update
	 */
	Map <K, V> insert(K k, V v);
	
	/**
	 * insert a key value pair,
	 * if there is already an entry, the old and new values
	 * are combinde by the combine operator
	 */
	Map <K, V> insertWith(K k, V v, final Fct2<V,V,V> combine);
	
	/**
	 * modify an an attribute value for a given key,
	 * if the key isn't found in the map, the map remains unchanged.
	 * 
	 * @param k the key for which the attribute is to be modified
	 * @param adj the modification function
	 * @return the modified map
	 */
	Map <K, V> adjust(K k, final Fct<V,V> adj);
	
	/**
	 * remove an entry from a map,
	 * if key is not found, the map remains unchanged
	 */
	Map <K, V> remove(K k);
	
	/**
	 * select an arbitray key value pair from a map,
	 * if the map is empty, the null reference is returned
	 */
	Pair <K,V> choose();
	
	/**
	 * union of two maps.
	 * If the maps contain equal keys, the attribute value of the 2. map
	 * is taken as attribute value for the result
	 */
	Map <K, V> union(final Map <K, V> m2);

	/**
	 * remove all keys from a map that occure in another map
	 */
	Map <K, V> difference(final Map <K, V> m2);
	
	/**
	 * restrict a map to a set of keys given by another map
	 * @param m2
	 */
	Map <K, V> intersection(final Map <K, V> m2);
	
	/**
	 * merge two maps with a combining function for vor the values for
	 * key occuring in both maps
	 */
	Map <K, V> unionWith(final Map <K, V> m2, final Fct2<V,V,V> f2);

	/**
	 * build a map with the set of keys occuring in both arguments
	 * by combining the values
	 */
	Map <K, V> intersectionWith(final Map <K, V> m2, final Fct2<V,V,V> f2);

	/** 
	 * @param proto acts as prototype for construction of the result set.
	 * The result set is implemented by the same class as proto
	 * @return all keys in the map
	 */
	Set <K> keys(Set<K> proto);
	
	/**
	 * @param proto acts as prototype for the construction of the result set
	 * @return all values in the map as a set
	 */
	Set <V> values(Set<V> proto);
	
	/**
	 * 
	 * @param mf the value processing function
	 * @return a map with the same set of keys but the new attribute values computed by
	 * mf
	 */
	Map <K, V> map(final Fct<V,V> mf);
	
	/**
	 * 
	 * @param <V2> the value type of the result map
	 * @param mf the value processing function
	 * @param proto acts as prototype for construction of the result set.
	 * @return a map with the same set of keys but the new attribute values computed by
	 * mf
	 */
	<V2>
	Map <K, V2> map(final Fct<V, V2> mf, Map<K,V2> proto);
	
	/**
	 * process all e
	 * @param <R> the type of the result
	 * @param f the combining function as an object
	 * @param init the initial value for the combining function,
	 * this is the result for the empty map
	 * @return the result
	 */
	<R>
	R fold(final Fct3<K,V,R,R> f, R init);
	
	/**
	 * process all entries of a map and combine the results
	 * see foldMap and fold1Map in Set interface for further documentation
	 * @param <R> the result type
	 * @param combine the result combination function
	 * @param map the mapping of the key value pairs
	 * @param init the initial value for the combine function
	 */
	<R>
	R foldMap(final Fct2<R,R,R> combine, final Fct2<K,V,R> map, R init);
	
	/**
	 * process all entries of a none empty map
	 * @param <R> the type of the result
	 * @param combine the combining function
	 * @param map the mapping function from key and value to res
	 * @param emptyRes the result for the empty map
	 * @return the result
	 */
	<R>
	R fold1Map(final Fct2<R,R,R> combine, Fct2<K, V, R> map, R emptyRes);

	/**
	 * test, whether for all key value pairs in a map a predicate p holds,
	 * generalisation of logical and
	 */
	Boolean all(Fct2<K,V,Boolean> p);
	
	/**
	 * test, whether for some key value pairs in a map a predicate p holds,
	 * generalisation of logical or
	 */
	Boolean any(Fct2<K,V,Boolean> p);
	
	/**
	 * conversion into a list of pairs
	 */
	List<Pair<K,V>> toList(List<Pair<K,V>> proto);
}