/**
 * $Id: MapImpl.java,v 1.7 2006/02/24 21:21:02 uwe Exp $
 */
package de.fhw.data.map;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Fct3;
import de.fhw.data.function.Predef;
import de.fhw.data.set.Set;
import de.fhw.data.tuple.Pair;
import de.fhw.data.list.*;

/**
 * All default implementations for maps.
 * Most of these implementations use folds,
 * so the methods make a full traversal of the maps
 * This leeds to extremly inefficient operations.
 * 
 * @author uwe
 *
 * @param <K> the Key type
 * @param <V> the Value type
 */
public abstract
class MapImpl<K, V>
implements Map<K,V> {
	/**
	 * isEmpty() is implemented by card() == 0
	 */
	public
	boolean isEmpty() {
		return
		card() == 0;
	}
	
	/**
	 * card is implemented with fold and counting
	 */
	public
	int card() {
		return
		fold(new Fct3<K,V,Integer,Integer>() {
			public
			Integer apply(K k, V v, Integer r) {
				return
				r + 1;
			}
		}
		, 0
		);
	}
	
	/**
	 * isIn is implemented with getValue and a null test
	 */
	public
	boolean contains(K k) {
		return
		getValue(k) != null;
	}

	/**
	 * isEqualTo implemented with card and all
	 */
	public
	boolean isEqualTo(final Map<K,V> m2) {
		if (this == m2)
			return true;
		
		if (card() != m2.card())
			return false;
		
		Fct2<K,V,Boolean> inMap2 =
			new Fct2<K,V,Boolean>() {
			public
			Boolean apply(K k, V v) {
				V v2 = m2.getValue(k);
				return
				v2 != null
				&&
				v.equals(v2);
			}
		};
		return
		all(inMap2);
	}
	
	/**
	 * getValue is implemented with fold and a compare,
	 * null is returned, when the key is not contained in the map
	 */
	public
	V getValue(final K k) {
		return
		fold(new Fct3<K,V,V,V>() {
			public
			V apply(K k1, V v, V r) {
				if (r == null && k.equals(k1))
					return
					v;
				return
				r;
			}
		}
		,(V)null
		);
	}

	public
	V at(final K k) {
		V res = getValue(k);
		if (res == null)
			throw
			new IndexOutOfBoundsException("at: key not found in map, key = " + k.toString());
		return res;
	}

	public
	Pair <K,V> choose() {
		Pair <K,V> res = choose1();
		if (res == null)
			throw
			new IllegalArgumentException("choose applied to empty set");	
		return res;
	}

	protected
	abstract
	Pair <K,V> choose1();
	
	/**
	 * insertWith is implemented with getValue and insert
	 */
	public
	Map <K, V> insertWith(K k, V v, final Fct2<V,V,V> combine) {
		V v2 = getValue(k);
		
		if (v2 == null)
			return
			insert(k,v);
		return
		insert(k, combine.apply(v2,v));
	}
	
	/**
	 * adjust is implemented with getValue and insert
	 */
	public
	Map <K, V> adjust(K k, final Fct<V,V> adj) {
		V v = getValue(k);
		
		if (v == null)
			return
			this;
		return
		insert(k, adj.apply(v));
	}
	
	/**
	 * remove is implemented with fold, empty and insert
	 */
	public
	Map <K, V> remove(final K k) {
		return
		fold(new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map <K, V> apply(K k1, V v, Map <K, V> r) {
				if (k.equals(k1))
					return
					r;
				return
				r.insert(k1,v);
			}
		}
		,empty()
		);
	}
	
	/**
	 * union is implemented with unionWith
	 */
	public
	Map <K, V> union(Map <K, V> m2) {
		return
		unionWith(m2,
				new Fct2<V,V,V>(){
			public
			V apply(V x, V y) {
				return x;
			}
		}
		);
	}
	
	/**
	 * difference is implemented with fold, isIn and insert
	 */
	public
	Map <K, V> difference(final Map <K, V> m2) {
		return
		fold(new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map<K, V> apply(K k, V v, Map<K, V> r) {
				if (m2.contains(k))
					return r;
				return
				r.insert(k,v);
			}
		}
		,empty()
		);
	}
	
	/**
	 * intersection is implemented with fold, isIn and insert
	 */
	public
	Map <K, V> intersection(final Map <K, V> m2) {
		return
		fold(new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map<K, V> apply(K k, V v, Map<K, V> r) {
				if (m2.contains(k))
					return r.insert(k,v);
				return
				r;
			}
		}
		,empty()
		);
	}
	
	/**
	 * unionWith is implemented with fold and insert
	 */
	public
	Map <K, V> unionWith(Map <K, V> m2, final Fct2<V,V,V> f2) {
		return
		fold(new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map<K, V> apply(K k, V v, Map<K, V> r) {
				return
				r.insertWith(k,v,f2);
			}
		}
		,m2
		);
	}
	
	
	/**
	 * intersectionWith is implemented with fold and insert
	 */
	public
	Map <K, V> intersectionWith(final Map <K, V> m2, final Fct2<V,V,V> f2) {
		return
		fold(new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map<K, V> apply(K k, V v, Map<K,V> r) {
				V v2 = m2.getValue(k);				
				return
				v2 == null
				? r
						: r.insert(k,f2.apply(v,v2));
			}
		}
		,m2
		);
	}
	
	/**
	 * implementation of the simple map function
	 * with fold
	 */
	public
	Map <K, V> map(final Fct<V,V> mf) {
		
		Fct3<K,V,Map<K,V>,Map<K,V>> f
		= new Fct3<K,V,Map<K,V>,Map<K,V>>() {
			public
			Map<K,V> apply(K k, V v, Map<K,V> res) {
				return
				res.insert(k,mf.apply(v));
			}
		};
		// return map(mf,this); is ambiguous (why???)
		return
		fold(f,empty());
	}
	
	/**
	 * implementation of the general map function with fold
	 */
	public
	<V2>
	Map <K,V2> map(
			final Fct<V, V2> mf,
			Map <K, V2> proto) {
		
		Fct3<K,V,Map<K,V2>,Map<K,V2>> f
		= new Fct3<K,V,Map<K,V2>,Map<K,V2>>(){
			public
			Map <K, V2> apply(K k, V v,	Map <K, V2> res) {
				return
				res.insert(k,mf.apply(v));
			}
		};
		return
		fold(f,proto.empty());
	}
	
	/**
	 * compute all keys
	 * 
	 */
	public
	Set<K> keys(Set<K> acc) {
		return
		fold(new Fct3<K,V,Set<K>,Set<K>>(){
			public
			Set<K> apply(K k, V v, Set<K> r) {
				return
				r.insert(k);
			}
		}
		,acc.empty()
		);
	}
	/**
	 * compute all attribute values in a set
	 */
	public
	Set<V> values(Set<V> acc) {
		return
		fold(new Fct3<K,V,Set<V>,Set<V>>(){
			public
			Set<V> apply(K k, V v, Set<V> r) {
				return
				r.insert(v);
			}
		}
		,acc.empty()
		);
	}

	/**
	 * all quantor implemented with fold
	 */
	public
	Boolean all(Fct2<K,V,Boolean> p) {
		return
		foldMap(Predef.Bool.and,p,true);
	}
	
	/**
	 * exists quantor implemented with fold
	 */
	public
	Boolean any(Fct2<K,V,Boolean> p) {
		return
		foldMap(Predef.Bool.or,p,false);
	}
	
	/**
	 * toList conversion with fold
	 */
	public
	List<Pair<K,V>> toList(List<Pair<K,V>> proto) {
		Fct3<K,V,List<Pair<K,V>>,List<Pair<K,V>>> cv =
			new Fct3<K,V,List<Pair<K,V>>,List<Pair<K,V>>>() {
			public
			List<Pair<K,V>> apply(K k, V v, List<Pair<K,V>> res) {
				return res.cons(new Pair<K,V>(k,v));
			}
		};
		return fold(cv,proto.empty());
	}
	
	/**
	 * equals implemented with isEqualTo
	 */
	@Override
	public
	boolean equals(Object o2) {
		if (o2 instanceof Map)
			return this.isEqualTo((Map<K,V>)o2);
		return false;
	}

	/**
	 * toString is implemented with isEmpty and fold1
	 */
	@Override
	public
	String toString() {
		if (isEmpty())
			return "{}";
		Fct2<K,V,String> kvShow =
			new Fct2<K,V,String>() {
			public
			String apply(K k, V v) {
				return
				k.toString() + ":->" + v.toString();
			}
		};
		return
		"{"
		+ fold1Map(Predef.Str.join(", "), kvShow,"")
		+ "}";
	}
	
}
