/**
 * $Id: SetImpl.java,v 1.8 2006/07/07 11:38:55 uwe Exp $
 */
package de.fhw.data.set;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Predef;
import de.fhw.data.list.*;

/**
 * All default implementations for sets.
 * Most of these implementations use folds,
 * so the methods make a full traversal of the sets
 * This leeds to extremly inefficient operations.
 * 
 * @param <E> the Element type
 */

public abstract
class SetImpl<E>
implements Set<E> {
	/**
	 * @return a single element set
	 */
	public
	Set <E> single(E e) {
		return empty().insert(e);
	}
	/**
	 * isEmpty() is implemented by card() == 0
	 */
	public
	boolean isEmpty() {
		return
		card() == 0;
	}

	/**
	 * isSubset implemented with all
	 */
	public
	boolean isSubsetOf(final Set<E> s2){
		Fct<E,Boolean> isElem =
		new Fct<E,Boolean>() {
			public
			Boolean apply(E e) {
				return
				s2.contains(e);
			}
		};
		return all(isElem);
	}

	/**
	 * isEqualTo implemented with card and contains
	 * 
	 */
	public
	boolean isEqualTo(Set<E> s2) {
		if (this == s2)
			return true;
		
		if (card() != s2.card())
			return false;
		
		return	contains(s2);
	}
	
	/**
	 * card is implemented with fold and counting
	 */
	public
	int card() {
		return
		fold(new Fct2<E,Integer,Integer>() {
			public
			Integer apply(E e, Integer r) {
				return
				r + 1;
			}
		}
		, 0
		);
	}
	
	/**
	 * contains is implemented with getValue and a null test
	 */
	public
	boolean contains(final E e) {
		return
		fold(new Fct2<E,Boolean,Boolean>() {
			public
			Boolean apply(E e1, Boolean r) {
				if (r)
					return r;
				return
				e.equals(e1);
			}
		}
		,(Boolean)false
		);
	}

	/**
	 * contains implemented with isSubsetOf
	 */
	public
	boolean contains(Set<E> s2) {
		return
		s2.isSubsetOf(this);
	}

	public E choose() {
		E e = choose1();
		if (e == null)
			throw
			new IllegalArgumentException("choose applied to empty set");	
		return e;
	}

	protected
	abstract
	E choose1();
	
	/**
	 * findMin implementation via fold
	 */
	public
	E findMin() {
		throw
		new UnsupportedOperationException("findMin not supported for Set implementation");
	}
	/**
	 * findMax implementation via fold
	 */
	public
	E findMax() {
		throw
		new UnsupportedOperationException("findMax not supported for Set implementation");
	}

	/**
	 * remove the minimum element
	 * 
	 * raise UnsupportedOperationException
	 */
	public
	Set <E> removeMin() {
		throw
		new UnsupportedOperationException("removeMin not supported for Set implementation");
	}
	
	/**
	 * remove the maximum element
	 * 
	 * raise UnsupportedOperationException
	 */
	public
	Set <E> removeMax() {
		throw
		new UnsupportedOperationException("removeMax not supported for Set implementation");
		
	}
	
	/**
	 * remove is implemented with fold, empty and insert
	 */
	public
	Set <E> remove(final E e) {
		return
		fold(new Fct2<E,Set<E>,Set<E>>() {
			public
			Set <E> apply(E e1, Set <E> r) {
				if (e.equals(e1))
					return
					r;
				return
				r.insert(e1);
			}
		}
		,empty()
		);
	}
	
	/**
	 * union is implemented with unionWith
	 */
	public
	Set <E> union(Set <E> s2) {
		return
		fold(new Fct2<E,Set<E>,Set<E>>() {
			public
			Set<E> apply(E e, Set<E> r) {
				return
				r.insert(e);
			}	
		}
		,s2);
	}
	
	/**
	 * difference is implemented with fold, contains and insert
	 */
	public
	Set <E> difference(final Set <E> s2) {
		return
		fold(new Fct2<E,Set<E>,Set<E>>() {
			public
			Set<E> apply(E e, Set<E> r) {
				if (s2.contains(e))
					return r;
				return
				r.insert(e);
			}
		}
		,empty()
		);
	}

	/**
	 * intersection is implemented with fold, contains and insert
	 */
	public
	Set <E> intersection(final Set <E> s2) {
		return
		fold(new Fct2<E,Set<E>,Set<E>>() {
			public
			Set<E> apply(E e, Set<E> r) {
				if (s2.contains(e))
					return r.insert(e);
				return
				r;
			}
		}
		,empty()
		);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * fold implemented with choose and remove,
	 * if this is an appropriate implementation,
	 * call it in the concrete implementation for fold.
	 * 
	 * @see #eclipseBugFoldMap(Fct2, Fct, R)
	 */
	protected
	<R>
	R eclipseBugFold(final Fct2<E,R,R> f, R init) {
		if (isEmpty())
			return init;
		E e1 = choose();
		R r1 = remove(e1).fold(f,init);
		return f.apply(e1,r1);
	}
	
	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Eclipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, rename this to foldMap and make it public,
	 * foldMap implementaions in subclasses, which call this method
	 * can then be removed
	 */
	protected
	<R>
	R eclipseBugFoldMap(final Fct2<R,R,R> combine, final Fct<E,R> map, R init) {
		Fct2<E,R,R> foldf =
			new Fct2<E,R,R>() {
			public
			R apply(E e, R r) {
				return
				combine.apply(map.apply(e),r);
			}
		};
		return
		this.fold(foldf,init);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Eclipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, rename this to fold1Map and make it public,
	 * fold1Map implementaions in subclasses, which call this method
	 * can then be removed
	 */
	protected
	<R>
	R eclipseBugFold1Map(
			final Fct2<R,R,R> combine,
			final Fct<E,R> map,
			R emptyRes) {
		if (isEmpty())
			return emptyRes;
		
		E e = choose1();
		
		Fct2<E,R,R> foldf	=
			new Fct2<E,R,R>() {
			public
			R apply(E e, R r) {
				return
				combine.apply(map.apply(e),r);
			}
		};
		return
		remove(e).fold(foldf,map.apply(e));
	}

	/**
	 * general implementation of the simple map function
	 * with fold
	 */
	public
	Set<E> map(final Fct<E,E> mf) {
		
		Fct2<E,Set<E>,Set<E>> f
		= new Fct2<E,Set<E>,Set<E>>() {
			public
			Set<E> apply(E e, Set<E> res) {
				return
				res.insert(mf.apply(e));
			}
		};
		return
		fold(f,empty());
	}

	/**
	 * general implementation of the more flexible map function
	 * with fold
	 */
	public
	<E2>
	Set<E2> map(final Fct<E,E2> mf, Set<E2> proto) {
		
		Fct2<E,Set<E2>,Set<E2>> f
		= new Fct2<E,Set<E2>,Set<E2>>() {
			public
			Set<E2> apply(E e, Set<E2> res) {
				return
				res.insert(mf.apply(e));
			}
		};
		return
		fold(f,proto.empty());
	}
	
	/**
	 * filter implemented with fold and insert
	 */
	public
	Set <E> filter(final Fct<E,Boolean> filter) {
		Fct2<E,Set<E>,Set<E>> f
		= new Fct2<E,Set<E>,Set<E>>() {
			public
			Set<E> apply(E e, Set<E> res) {
				return
				filter.apply(e)
				? res.insert(e) : res;
			}
		};
		return
		fold(f,empty());
	}
	
	/**
	 * all quantor implemented with fold
	 */
	public
	Boolean all(Fct<E,Boolean> p) {
		return
		foldMap(Predef.Bool.and,p,true);
	}
	
	/**
	 * exists quantor implemented with fold
	 */
	public
	Boolean any(Fct<E,Boolean> p) {
		return
		foldMap(Predef.Bool.or,p,false);
	}

	/**
	 * toList conversion with fold
	 */
	public
	List<E> toList(List<E> proto) {
		Fct2<E,List<E>,List<E>> cv =
			new Fct2<E,List<E>,List<E>>() {
			public
			List<E> apply(E e, List<E> res) {
				return res.cons(e);
			}
		};
		return fold(cv,proto.empty());
	}
	
	@Override
	public
	boolean equals(Object o2) {
		if (o2 instanceof Set)
			return this.isEqualTo((Set<E>)o2);
		return false;
	}
	
	/**
	 * toString is implemented with isEmpty and fold1Map
	 */
	@Override
	public
	String toString() {
		if (isEmpty())
			return "{}";
		return
		"{"
		+ fold1Map(Predef.Str.join(", "), Predef.<E>show(), "")
		+ "}";
	}
}
